/****************************************************************************

  This file is part of the VRender library.
  Copyright (C) 2005 Cyril Soler (Cyril.Soler@imag.fr)
  Version 1.0.0, released on June 27, 2005.

  http://artis.imag.fr/Members/Cyril.Soler/VRender

  VRender is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.
  
  VRender is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with
  VRender; if not, write to the Free Software Foundation, Inc., 59 Temple
  Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/

#ifndef _VRENDER_NVECTOR3_H
#define _VRENDER_NVECTOR3_H

#include <iostream>
#include <stdexcept>

#include "Vector3.h"

namespace vrender
{
	template<class T> class T_NVector3
	{
		public:
			T_NVector3();
			T_NVector3(const T_NVector3<T>& u);
			inline T_NVector3(T x,T y,T z,bool normalization=true) 
			{
				setXYZ(x,y,z,normalization);
			}
			inline T_NVector3(const T_Vector3<T> &u,bool normalization=true) {setXYZ(u[0],u[1],u[2],normalization);}

			inline T x() const {return _n[0];}
			inline T y() const {return _n[1];}
			inline T z() const {return _n[2];}
			void setXYZ(T x,T y,T z,bool normalization=true);

			T_NVector3<T>& operator=(const T_NVector3<T>& u);

			inline friend bool operator==(const T_NVector3<T> &u,const T_Vector3<T>  &v) {return u.isEqualTo(v);}
			inline friend bool operator==(const T_Vector3<T>  &u,const T_NVector3<T> &v) {return v.isEqualTo(u);}
			inline friend bool operator==(const T_NVector3<T> &u,const T_NVector3<T> &v) {return u.isEqualTo(v);}
			inline friend bool operator!=(const T_NVector3<T> &u,const T_Vector3<T>  &v) {return !(u == v);}
			inline friend bool operator!=(const T_Vector3<T>  &u,const T_NVector3<T> &v) {return !(u == v);}
			inline friend bool operator!=(const T_NVector3<T> &u,const T_NVector3<T> &v) {return !(u == v);}
#ifdef A_VIRER
			bool isEqualTo(const T_NVector3<T>& u,T epsilon=T_epsilon) const;
			bool isEqualTo(const T_Vector3<T>&  u,T epsilon=T_epsilon) const;
#endif

			inline friend T_NVector3<T> operator-(const T_NVector3<T> &u)                   {return T_NVector3<T>(-u[0],-u[1],-u[2],false);}
			inline friend T_Vector3<T>  operator+(const T_NVector3<T> &u,const T_Vector3<T>  &v) {return T_Vector3<T>(u[0]+v[0],u[1]+v[1],u[2]+v[2]);}
			inline friend T_Vector3<T>  operator+(const T_Vector3<T>  &u,const T_NVector3<T> &v) {return T_Vector3<T>(u[0]+v[0],u[1]+v[1],u[2]+v[2]);}
			inline friend T_Vector3<T>  operator+(const T_NVector3<T> &u,const T_NVector3<T> &v) {return T_Vector3<T>(u[0]+v[0],u[1]+v[1],u[2]+v[2]);}
			inline friend T_Vector3<T>  operator-(const T_NVector3<T> &u,const T_Vector3<T>  &v) {return T_Vector3<T>(u[0]-v[0],u[1]-v[1],u[2]-v[2]);}
			inline friend T_Vector3<T>  operator-(const T_Vector3<T>  &u,const T_NVector3<T> &v) {return T_Vector3<T>(u[0]-v[0],u[1]-v[1],u[2]-v[2]);}
			inline friend T_Vector3<T>  operator-(const T_NVector3<T> &u,const T_NVector3<T> &v) {return T_Vector3<T>(u[0]-v[0],u[1]-v[1],u[2]-v[2]);}
			inline friend T     operator*(const T_NVector3<T> &u,const T_Vector3<T>  &v) {return u[0]*v[0] + u[1]*v[1] + u[2]*v[2];}
			inline friend T     operator*(const T_Vector3<T>  &u,const T_NVector3<T> &v) {return u[0]*v[0] + u[1]*v[1] + u[2]*v[2];}
			inline friend T     operator*(const T_NVector3<T> &u,const T_NVector3<T> &v) {return u[0]*v[0] + u[1]*v[1] + u[2]*v[2];}
			inline friend T_Vector3<T>  operator*(T r,const T_NVector3<T> &u)            {return T_Vector3<T>(r*u[0],r*u[1],r*u[2]);}
			inline friend T_Vector3<T>  operator/(const T_NVector3<T> &u,T r)            {return T_Vector3<T>(u[0]/r,u[1]/r,u[2]/r);}

			inline friend T_Vector3<T>  operator^(const T_NVector3<T> &u,const T_Vector3<T>  &v) 
			{
				return T_Vector3<T>(	u[1]*v[2]-u[2]*v[1],
						u[2]*v[0]-u[0]*v[2],
						u[0]*v[1]-u[1]*v[0]);
			}

			inline friend T_Vector3<T>  operator^(const T_Vector3<T>  &u,const T_NVector3<T> &v) 
			{
				return T_Vector3<T>(	u[1]*v[2]-u[2]*v[1],
						u[2]*v[0]-u[0]*v[2],
						u[0]*v[1]-u[1]*v[0]);
			}
			inline friend T_NVector3<T> operator^(const T_NVector3<T> &u,const T_NVector3<T> &v) 
			{
				return T_Vector3<T>(	u[1]*v[2]-u[2]*v[1],
						u[2]*v[0]-u[0]*v[2],
						u[0]*v[1]-u[1]*v[0]);
			}

			inline T norm() const {return 1.0;}
			inline T squareNorm() const {return 1.0;}
			template<class U> friend std::ostream& operator<<(std::ostream &out,const T_NVector3<U> &u);

			T operator[](int i) const 
			{
				if((i < 0)||(i > 2))
					throw std::runtime_error("Out of bounds in T_NVector3<T>::operator[]") ;

				return _n[i];
			}

		private:
			void normalize();

			T _n[3];  //!< normalized vector

	}; // interface of T_NVector3


	// -------------------------------------------------------------------------- //
	//                                                                            //
	//  CONSTANT DEFINITIONS                                                      //
	//                                                                            //
	// -------------------------------------------------------------------------- //

	// -------------------------------------------------------------------------- //
	//                                                                            //
	//  CLASS IMPLEMENTATION                                                      //
	//                                                                            //
	// -------------------------------------------------------------------------- //

	// -----------------------------------------------------------------------------
	//! Default constructor (the default normalized vector is (1,0,0))
	template<class T> T_NVector3<T>::T_NVector3()
	{
		_n[0] = 1.0;
		_n[1] = 0.0;
		_n[2] = 0.0;
	}

	// -----------------------------------------------------------------------------
	//! Copy constructor
	template<class T> T_NVector3<T>::T_NVector3(const T_NVector3<T>& u)
	{
		_n[0] = u._n[0] ;
		_n[1] = u._n[1] ;
		_n[2] = u._n[2] ;
	}
	// -----------------------------------------------------------------------------
	//! Writing X,Y and Z coordinates
	template<class T> void T_NVector3<T>::setXYZ(T x,T y,T z,bool normalization)
	{
		_n[0] = x;
		_n[1] = y;
		_n[2] = z;
		if ( normalization ) normalize();
	}

	// -----------------------------------------------------------------------------
	//! Assignment
	template<class T> T_NVector3<T>& T_NVector3<T>::operator=(const T_NVector3<T>& u)
	{
		if ( &u != this )
		{
			_n[0] = u[0];
			_n[1] = u[1];
			_n[2] = u[2];
		}
		return *this;
	}
	// -----------------------------------------------------------------------------
	//! Out stream override: prints the 3 normalized vector components
	template<class T> std::ostream& operator<<(std::ostream& out,const T_NVector3<T>& u)
	{
		out << u._n[0] << " " << u._n[1] << " " << u._n[2];
		return out;
	}

	// -----------------------------------------------------------------------------
	//! Normalization
	//! Private method to do normalization (using Norm() method of the Vector class)
	//! when it is necessary (construction of a normalized vector for exemple).
	template<class T> void T_NVector3<T>::normalize()
	{
		T n = _n[0]*_n[0]+_n[1]*_n[1]+_n[2]*_n[2] ;

		if ( n > 0.0 )
		{
			_n[0] /= n;
			_n[1] /= n;
			_n[2] /= n;
		}
		else
			throw std::runtime_error("Attempt to normalize a null 3D vector.") ;
	}
}

#endif // _NVECTOR3_H
